{ Free Pascal unit to control the Music On Console Player (mocp)
  Version: 0.2
  Date: 2018-05-11
  Author: Michel Deslierres
  License: BDS style }

unit mocpi;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, process, strutils, fileutil;

const
  MOC_POLLING_WAIT_MS = 500;
  PLAY_TIMEOUT = 5000;

{ These are the basic commands to play Internet radio using
  the moc daemon.

  These are non blocking procedures that do not return any
  status information, mostly because MOC does not return
  information.}

procedure play(uri: string);
procedure pause;
procedure resume;
procedure stop;
procedure setVolume(vol: integer; relative: boolean = true);

{ These function query MOC to get the state, the current URI
  and songtitle or title }
function state: string;
function uri: string;
function title: string;

{ This function queries MOC using a Unix Domain Socket and
   returns the current volume. If an error occurs it
   returns -1}
function volume: integer;

{ This function returns the raw information string from the
  MOC daemon. It is available for debugging otherwise
  it could be confined to the implementation part. }
function  info: string;

{ These are blocking functions that either return when the stream has been
  found and will play or when the timeout is reached.

  Returns 0 if successul, -1 if the uri could not be resolved within the
  timeout period.}
function playwt(uri: string; timeout: integer = PLAY_TIMEOUT): integer; overload;

{ Returns the index of the first uri that could be resolved within the
  specified timeout period or else returns -1 }
function playwt(uris: array of string; timeout: integer = PLAY_TIMEOUT): integer; overload;

{ Plays a file and resumes ('nr) playing whatever it was playing before. }
procedure playnr(uri: string);


{ This unit starts a thread that polls the MOC daemon at regular
  intervals. If there is a relevant change in the status then
  all registered TMocStateChangedEvent handlers will be called.
}
type
  TMocEvent = (meState, meUri, meTitle);
  TMocEvents = set of TMocEvent;

  TMocState = record
    state: string;
    uri: string;
    title: string;
  end;

  { Event handler type, Events is a set of TMocEvent indicating elements
    changed in the MOC status. If meState is included in Events then
    the MOC state has changed since from the previous poled by the thread.
    Similarly, if meUri is in Events then MOC is now playing a different
    audio file/stream. If meTitle is in Events then the song title or
    title of the audio file/stream played by MOC has changed.
    The current MOC state, uri and title/songtitle are returnd in
    MocState. If songtitle is empty, then title is returned.}
  TMocStateChangedEvent = procedure(Sender: TObject; Events: TMocEvents;
                                    const MocState: TMocState) of object;

  { TMocPollingThread }

  TMocPollingThread = class(TThread)
  private
    FEvents: TMocEvents;
    FCurrentState: TMocState;
    FPolling: boolean;
    FPollingWait: integer;
    FListeners: array of TMocStateChangedEvent;
    procedure SetPollingWait(value: integer);
  protected
    procedure ReportChange;
    procedure Execute; override;
  public
    Constructor Create(CreateSuspended : boolean);
    procedure Pause;
    procedure Resume;
    function Paused: boolean;
    function AddListener(aHandler: TMocStateChangedEvent): integer;
    function RemoveListener(aHandler: TMocStateChangedEvent): integer;
    property PollingWait: integer read FPollingWait write SetPollingWait;
  end;

var
  MocPoller: TMocPollingThread;

implementation

uses
  ssockets; // only for function volume: integer;

function exc(cmd: array of string): string;
begin
  RunCommand('mocp', cmd, result, [poWaitOnExit, poNoConsole])
end;

procedure play(uri: string);
begin
  exc(['-l', uri]);
end;

procedure pause;
begin
  exc(['-P'])
end;

procedure resume;
begin
  exc(['-U'])
end;

procedure stop;
begin
  exc(['-s'])
end;

procedure setVolume(vol: integer; relative: boolean = true);
var
  str: string;
begin
  str := inttostr(vol);
  if relative then begin
    if vol > 0 then
      str := '+' + str
    else if vol = 0 then
      exit;
  end
  else if vol < 0 then
    vol := 0
  else if vol > 100 then
    vol := 100;
  exc(['-v', str])
end;

function  info: string;
begin
  result := exc(['-i'])
end;

function state: string;
begin
  // moc (or exc) appends a trailing #10
  result := trim(exc(['-Q', '%state']));
  if result = '' then
    result := 'NO SERVER';
end;

function uri: string;
begin
  result := trim(exc(['-Q', '%file']))
end;

function title: string;
var
  ss: string;
  p, n: integer;
begin
  ss := trim(exc(['-Q', '%title\n%song']));
  n := length(ss);
  p := pos(#10, ss);
  if p >= n then
    result := copy(ss, 1, p-1)
  else
    result := copy(ss, p+1, n-p);
end;

function volume: integer;
{ source python-moc fork by Michał Bielecki (y0g1)
  https://github.com/y0g1/python-moc

  def get_volume():
      s = socket.socket( socket.AF_UNIX, socket.SOCK_STREAM )
      s.connect(os.path.expanduser(Cli.socketfile))
      s.send(struct.pack('i', 0x1a))
      unpacker = struct.Struct('i i')
      data = s.recv(unpacker.size)
      s.close()
      return unpacker.unpack(data)[1]

      line 233 in
      https://github.com/y0g1/python-moc/blob/master/moc/__init__.py
}

const
  EV_DATA = 6;
  CMD_GET_MIXER: longint = $1A;
var
  sock: TUnixSocket;
  inbuf: array[0..1] of longint;
begin
  sock := TUnixSocket.create(expandfilename('~/')+'.moc/socket2');
  try
    sock.WriteBuffer(CMD_GET_MIXER, sizeof(CMD_GET_MIXER));
    fillchar(inbuf, sizeof(inbuf), 0);
    if (sock.read(inbuf, sizeof(inbuf)) = sizeof(inbuf))
    and (inbuf[0] = EV_DATA) then
      result := inbuf[1]
    else
      result := -1;
  finally
    sock.free;
  end;
end;

procedure GetMocState(var rec: TMocState);
{
State: PLAY
File: http://7emct0.akacast.akamaistream.net/7/282/177408/v1/rc.akacast.akamaistream.net/7EMCT0
Title:
Artist:
SongTitle:
Album:
CurrentTime: 00:00
CurrentSec: 0
Bitrate: 0kbps
AvgBitrate: 0kbps
Rate: 0kHz

State: PLAY
File: http://audio.wgbh.org:8000
Title: 89.7 WGBH | Boston
Artist:
SongTitle: 89.7 WGBH | Boston
Album:
}
var
  ss: string;
  p, q, n: integer;
begin
  ss := exc(['-Q', '%state\n%file\n%title\n%song']);
  n := length(ss);
  p := posEx(#10, ss, 1);
  rec.state := copy(ss, 1, p-1);
  if rec.state = '' then begin
    rec.state := 'NO SERVER';
    rec.uri := '';
    rec.title := '';
    exit;
  end;

  inc(p); // skip delim
  q := p;
  p := posEx(#10, ss, q);
  rec.uri := copy(ss, q, p-q);

  inc(p); // skip delim
  q := p;
  p := posEx(#10, ss, q);

  if p < n-1 then begin
    if p > 0 then
      q := succ(p);
    p := n+1;
  end
  else
    inc(p);
  rec.title := copy(ss, q, p-q-1)
end;

function playwt(uri: string; timeout: integer): integer;
begin
  result := playwt([uri], timeout);
end;

function playwt(uris: array of string; timeout: integer): integer;
var
  n, i: integer;
  oldPaused: boolean;

  function isPlaying: boolean;
  var
    t: integer;
  begin
    for t := 0 to n-1 do begin
      sleep(100);
    	result := state = 'PLAY';
      if result then exit;
    end;
  end;

begin
  if timeout < 100 then
    timeout := 100;
  n := timeout div 100;
  if timeout mod 100 <> 0 then
    inc(n);
  oldPaused := MocPoller.Paused;
  MocPoller.Pause;
  try
    for i := 0 to length(uris)-1 do begin
      play(uris[i]);
      if isPlaying then begin
        result := i;
        exit;
      end;
    end;
    result := -1;
  finally
    if oldPaused then
      MocPoller.Pause
    else
      MocPoller.Resume;
  end;
end;

procedure playnr(uri: string);
var
  oldPaused: boolean;
  oldurl: string;
  oldstatus: integer;
begin
  if not fileexists(uri) then exit;
  case state of
    'NO SERVER': exit;
    'PLAY': oldstatus := 1;
    'PAUSE': oldstatus := 2;
  else
     oldstatus := 0;
  end;
  if oldstatus > 0 then
    oldurl := mocpi.uri;
  oldPaused := MocPoller.Paused;
  MocPoller.Pause;
  try
    play(uri);
    repeat
      sleep(300);
    until state = 'STOP';
    if oldstatus > 0 then begin
      play(oldurl);
      if oldstatus = 2 then
        pause;
    end;
  finally
    if oldPaused then
      MocPoller.Pause
    else
      MocPoller.Resume;
  end;
end;

function TMocPollingThread.AddListener(aHandler: TMocStateChangedEvent): integer;
var
  n: integer;
begin
  n := length(FListeners);
  result := succ(n);
  setlength(FListeners, result);
  FListeners[n] := aHandler;
end;

constructor TMocPollingThread.Create(CreateSuspended : boolean);
begin
  inherited Create(CreateSuspended);
  FPollingWait := MOC_POLLING_WAIT_MS;
  FreeOnTerminate := True;
end;

procedure TMocPollingThread.Pause;
begin
  FPolling := false;
end;

procedure TMocPollingThread.Resume;
begin
  FPolling := true;
end;

function TMocPollingThread.Paused: boolean;
begin
  result := not FPolling
end;

procedure TMocPollingThread.Execute;
var
  NewState : TMocState;
begin
  while (not Terminated) do begin
    sleep(FPollingWait);
    if FPolling then begin
      GetMocState(NewState);
      FEvents := [];
      if NewState.state <> FCurrentState.state then
        include(FEvents, meState);

      if NewState.uri <> FCurrentState.uri then
        include(FEvents, meUri);

      if NewState.title <> FCurrentState.title then
        include(FEvents, meTitle);

      if FEvents <> [] then begin
        FCurrentState := NewState;
        Synchronize(@ReportChange);
      end;
    end;
  end;
end;

function TMocPollingThread.RemoveListener(aHandler: TMocStateChangedEvent): integer;
var
  i,j,newlen: integer;
begin
  newlen := length(FListeners);
  for i := length(FListeners)-1 downto 0 do begin
    if FListeners[i] = aHandler then begin
      for j := i to newlen-2 do
        FListeners[j] := FListeners[j+1];
      dec(newlen);
    end;
  end;
  if newlen < length(FListeners) then begin
    setlength(FListeners, newlen);
    result := newlen;
  end
  else
    result := -1;
end;

procedure TMocPollingThread.SetPollingWait(value: integer);
begin
  if value < 100 then
    value := 100
  else if value > 5*60*1000 then
    value := 5*60*1000;
  FPollingWait := value;
end;

procedure TMocPollingThread.ReportChange;
// this method is executed by the mainthread and can therefore access all GUI elements.
var
  mocState: TMocState;
  i: integer;
begin
  if length(FListeners) > 0 then begin
    mocState := FCurrentState;
    for i := 0 to length(FListeners)-1 do
      FListeners[i](self, FEvents, mocState);
  end;
end;

initialization
  MocPoller := TMocPollingThread.create(true);
  MocPoller.Start;
finalization
  MocPoller.Terminate;
  MocPoller.free;
end.

